#!/bin/bash
#***********************************************************
# makepatch                                                *
#   Make a patch script by substituting #include           *
#   directives in a shell script.                          *
#                                                          *
#   valid include statements are:                          *
#     #includedir                                          *
#     #includefile                                         *
#                                                          *
#   #includedir - makes a tarball of the dir and uuencodes *
#      it, then includes the uuencoded data in place of    *
#      the #includedir statement                           *
#   #includefile - takes the given file and uuencodes it,  *
#      then includes the uuencoded data in place of the    *
#      statement                                           *
#                                                          *
#  NOTE:                                                   *
#    It's best to supply a full path to both of these      *
#    directives, but they will try to assume relative      *
#    paths where possible.                                 *
#                                                          *
#***********************************************************

# needed shell options
shopt -s extglob        # extended pattern matching features of bash
shopt -s expand_aliases # expand aliases
shopt -s xpg_echo       # expand backslash sequences in echo lines -
                        # this way we don't have to supply '-e' to echo cmds

# Permit me the minor perl-ism...makes code so much cleaner
die () {
    local ex=$?
    echo "${1}"
    exit ${ex}
}

# initialize variables:
bzip=false
gzip=false
cnt=0

usage="\nUsage: `basename $0` [-h] [-v] [-(z|j)] [-f <filename>] <output file> \n\
  -v              Verbose Mode
  -f <filename>   The filename to take as input file\n\
  -z              use gzip for compression of tar'd dirs and files\n\
                  Otherwise, no compression is used
  -j              use bzip2 for compression of tar'd dirs and files\n\
                  Otherwise, no comression is used\n\
  -h              print this message\n"

# parse command line opts
while getopts :vf:zjh option
do
case "${option}" in
    v)
        verbose=true
    ;;
    f)  # Source file - script on which to do the replace functions
        test -f ${OPTARG} && infile=${OPTARG} || die "\n ${OPTARG} is not a valid file! \n"
    ;;
    z)  # use gzip for tar archives
		gzip=true
    ;;
    j)  # use bzip2 for tar archives
		bzip=true
    ;;
    h)  # help message
        echo "${usage}"
        exit 0
    ;;
    \?)  echo "${usage}"
        exit 1
    ;;
esac
done

test $# -lt 1 && die "$usage"
#shift the argument stack until we have the last one
until test $# -le 1; do shift; done 

#interpret our arguments:
$gzip && taropts="-z" && compress=gzip
$bzip && taropts="-j" && compress=bzip2

# sanity checkpoint:
test -n $infile || die "Need to specify a file to parse! (use '-f <filename>')"
test -n $outfile || die "Need to specify an output file!"
$bzip && $gzip && echo "Can't use gzip and bzip together...choose one!!!"
[[ "${1}" != @(-*) ]] && outfile=${1} || die "${1} is not a valid output filename"
test -f $infile || die "$infile does not exist!"
[[ "$outfile" != @(/*) ]] && outfile=$PWD/$outfile #if outfile is relative, make it not

# make sure output file is empty
> $outfile

li=0 #line counter
while read lineout; do
	let li++
	if [ "${lineout#\#include}" != "${lineout}" ]; then
		directive=${lineout% *}
		target=${lineout#* }
		case $directive in
		"#includedir")
			test -d ${target} || die "${target} is not a directory on line $li"
			echo "#beginning of ${directive} ${target}" >> ${outfile}
			echo "cd ${target%/*} &&" >> ${outfile}
			echo '{ uudecode <<!' >> ${outfile}
			test -d ${target%/*} && cd ${target%/*}; 
				tar -c ${taropts} ${target##*/} | uuencode -m /dev/stdout >> ${outfile}
			echo "!\n} | tar ${taropts} -x &>/dev/null" >> ${outfile}
			echo "#end of ${directive} ${target}" >> ${outfile}
		;;
		"#includefile")
			if [ "`file --mime $target`" == @($target: application/*) ]; then
				echo "#beginning of already compressed ${directive} ${target}" >> ${outfile}
				test -d ${target%/*} && echo "cd ${target%/*} &&" >> ${outfile}
				echo '{ uudecode <<!' >> ${outfile}
				test -d ${target%/*} && cd ${target%/*};  
					uuencode -m ${target##*/} /dev/stdout >> ${outfile}
				echo "!\n} >${target##*/}" >> ${outfile}
				echo "#end of already compressed ${directive} ${target}" >> ${outfile}
			else
				echo "#beginning of compressed ${directive} ${target}" >> ${outfile}
				test -d ${target%/*} && echo "cd ${target%/*} &&" >> ${outfile}
				echo '{ uudecode <<!' >> ${outfile}
				test -d ${target%/*} && cd ${target%/*}; 
					$compress -9c ${target##*/} | uuencode -m /dev/stdout >> ${outfile}
				echo "!\n} | $compress -d >${target##*/}" >> ${outfile}
				echo "#end of compressed ${directive} ${target}" >> ${outfile}
			fi
		;;
		esac
	else
		cat >> $outfile <<-!
		${lineout}
		!
	fi
done < <(sed 's!\\!\\\\!g' $infile)

# make the output file executable
chmod +x $outfile

